# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_loss.ipynb (unless otherwise specified).

__all__ = ['LossRouting', 'CombinedLoss']

# Cell
from fastai.basics import *

# Cell

from typing import Callable
from dataclasses import dataclass

@dataclass
class LossRouting:
    loss_func: Callable
    pred_idx: int
    target_idx: int
    weight: float = 1.0

class CombinedLoss:
    """Applies loss functions to multiple model outputs and sums them.
    If applicable, it can decode and compute activations for each model output."""

    def __init__(self, *loss_routings):
        self.loss_routings = loss_routings

    def __call__(self, outs, *targets, **kwargs):
        return sum([
            routing.weight*routing.loss_func(outs[routing.pred_idx], targets[routing.target_idx])
            for routing in self.loss_routings
        ])

    def activation(self, outs):
        return [getattr(routing.loss_func, 'activation', noop)(outs[routing.pred_idx]) for routing in self.loss_routings]

    def decodes(self, outs):
        return [getattr(routing.loss_func, 'decodes', noop)(outs[routing.pred_idx]) for routing in self.loss_routings]

    @classmethod
    def from_one_to_one_routing(cls, *loss_funcs):
        return cls(*[LossRouting(loss_func, pred_idx=i, target_idx=i, weight=1.0) for i, loss_func in enumerate(loss_funcs)])



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_metric.ipynb (unless otherwise specified).

__all__ = ['RoutedAccumMetric', 'route_to_metric', 'mtl_metrics']

# Cell
from fastcore.basics import GetAttr, store_attr, ifnone
from types import FunctionType
from fastai.metrics import Metric, AvgMetric, AccumMetric

# Cell

class _LearnerProxy(GetAttr):
    _default = 'learn'
    def __init__(self, learn, pred_idx, target_idx):
        store_attr()
        self.pred = self.learn.pred[pred_idx]
        self.y = self.learn.y[target_idx]


class RoutedAccumMetric(AccumMetric, GetAttr):
    "AccumMetric with predictions and targets for a specific model head."
    _default = 'metric'
    def __init__(self, metric, pred_idx, target_idx):
        self.metric = metric
        self._name = metric.name
        self.pred_idx = pred_idx
        self.target_idx = target_idx

    def reset(self):
        "Clear all targs and preds"
        return self.metric.reset()

    def accumulate(self, learn):
        "Store targs and preds from `learn`, using activation function and argmax as appropriate"
        return self.metric.accumulate(_LearnerProxy(learn, self.pred_idx, self.target_idx))

    def __call__(self, preds, targs):
        "Calculate metric on one batch of data"
        return self.metric(preds[self.pred_idx], targs[self.target_idx])

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

def route_to_metric(metric, pred_idx, target_idx):
    """Routes model output at `pred_idx` and target at index `target_idx` to metric"""
    if isinstance(metric, type):
        metric = metric()
    if isinstance(metric, FunctionType):
        func = lambda preds, *targs, **kwargs: metric(preds[pred_idx], targs[target_idx], **kwargs)
        func.__name__ = metric.__name__
        return AvgMetric(func)
    if isinstance(metric, Metric):
        return RoutedAccumMetric(metric, pred_idx, target_idx)
    raise ValueError("Unsupported metric type; must be either function or Metric")


# Cell

def mtl_metrics(*metrics_list):
    """Convenience function to route each prediction to list of metrics by their order."""
    return [route_to_metric(m, i, i) for i, metrics in enumerate(metrics_list) for m in metrics]


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"LossRouting": "01_loss.ipynb",
         "CombinedLoss": "01_loss.ipynb",
         "RoutedAccumMetric": "02_metric.ipynb",
         "route_to_metric": "02_metric.ipynb",
         "mtl_metrics": "02_metric.ipynb"}

modules = ["loss.py",
           "metric.py"]

doc_url = "https://bdsaglam.github.io/fastmtl/"

git_url = "https://github.com/bdsaglam/fastmtl/tree/master/"

def custom_doc_links(name): return None
